/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "fwk.h"
#include <iostream>

#if defined(WIN32)
#include <float.h>
#elif defined(__GNUC__)
#include <fenv.h>
#endif

#define ANSI_COLOR_YELLOW "\x1b[1;33m"
#define ANSI_COLOR_BLUE "\x1b[1;34m"
#define ANSI_COLOR_RESET "\x1b[0m"

FWK_API void testswig(bool i) // testswig(3.0) displays 1 (no error and cast float=>bool)
{
    std::cout << "testswig received i = " << i << '\n';
}

/**
 * @brief Enable Floating Point Exception
 *
 * Imported from Metafor
 * @authors Adrien Crovato
 */
FWK_API void enableFpe()
{
#if !defined(__MINGW32__) // MinGW: 2 "float.h" and the wrong one is included by default...
    std::cout << ANSI_COLOR_BLUE << "Enabling Floating-Point Exceptions..." << ANSI_COLOR_RESET << std::endl;
#if defined(WIN32)
    // doc https://msdn.microsoft.com/en-us/library/e9b52ceh(v=vs.110).aspx
    // _EM_INVALID | _EM_DENORMAL | _EM_ZERODIVIDE | _EM_OVERFLOW | _EM_UNDERFLOW | _EM_INEXACT
    _clear87();                                                                    // doc : "clearing the mask sets the exceptions"
    _control87(_EM_INVALID | _EM_DENORMAL | _EM_UNDERFLOW | _EM_INEXACT, _MCW_EM); // doc : "setting the _MCW_EM mask hides the exception"
#elif defined(__APPLE__)
    // we should try feraiseexcept (c++11)
#elif defined(__GNUC__)
    // FE_ALL_EXCEPT|FE_DIVBYZERO|FE_INEXACT|FE_INVALID|FE_OVERFLOW|FE_UNDERFLOW
    feenableexcept(FE_OVERFLOW | FE_DIVBYZERO | FE_INVALID);
#else // defined(WIN32)
    std::cout << ANSI_COLOR_YELLOW << "Floating-Point Exceptions not implemented for this compiler!" << ANSI_COLOR_RESET << std::endl;
#endif
#endif //! defined(__MINGW32__)
}

/**
 * @brief Disable Floating Point Exception
 * @authors Adrien Crovato
 */
FWK_API void disableFpe()
{
#if defined(__GNUC__) && !defined(__clang__) && !defined(__MINGW32__)
    std::cout << ANSI_COLOR_BLUE << "Disabling Floating-Point Exceptions..." << ANSI_COLOR_RESET << std::endl;
    fedisableexcept(FE_ALL_EXCEPT);
#else
    std::cout << "disableFpe() is not implemented" << std::endl;
#endif
}
