/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wTetra4.h"
#include "wSfTetra4.h"
#include "wGaussTetra4.h"
#include "wCacheTetra4.h"
#include "wNode.h"

using namespace tbox;

Tetra4::Tetra4(int n, Tag *_ptag, Tag *_etag,
               std::vector<Tag *> &_parts, std::vector<Node *> &nods) : Element(n, _ptag, _etag, _parts, nods)
{
}

/**
 * @brief Return cache handling shape functions and Gauss points for all Tetra4 elements
 */
Cache &Tetra4::getVCache() const
{
    return getCache(order);
}

/**
 * @brief Initialize precomputed values
 */
void Tetra4::initValues(bool isvol, int ordr)
{
    // Set flags
    isVol = isvol;
    order = ordr;
    hasVals = true;
    // Update values
    size_t ngp = getCache(order).getVGauss().getN();
    Js.resize(ngp);
    iJs.resize(ngp);
    detJs.resize(ngp);
    this->update();
}
/**
 * @brief Initialize precomputed gradients
 */
void Tetra4::initGradients()
{
    // Set flags
    hasGrads = true;
    // Update values
    size_t ngp = getCache(order).getVGauss().getN();
    gradJs.resize(ngp);
    gradDetJs.resize(ngp);
    this->update();
}
/**
 * @brief Update precomputed values and gradients
 */
void Tetra4::update()
{
    // Update gradients
    if (hasVals)
    {
        // Jacobian
        if (!isVol)
            throw std::runtime_error("Tetra4::update(): the problem must always have the same dimension as Tetra4!\n");
        // Tetra4 is always of same dimension as Problem (volume element)
        this->buildJ();
        // Volume and CG
        this->computeV();
        this->computeCg();
    }
    // Update gradients
    if (hasGrads)
    {
        // Jacobian
        if (!isVol)
            throw std::runtime_error("Tetra4::updateGradients(): the problem must always have the same dimension as Tetra4!\n");
        // Tetra4 is always of same dimension as Problem (volume element)
        this->buildGradientJ();
        // Volume
        this->buildGradientV(3);
    }
}

/**
 * @brief Compute Jacobian matrix, inverse and determinant
 *
 * J_ij(3,3) = diN_n * xj_n (n is the node)
 */
void Tetra4::buildJ()
{
    CacheTetra4 &cache = getCache(order);
    for (size_t k = 0; k < cache.getVGauss().getN(); ++k)
    {
        Eigen::MatrixXd const &dff = cache.getDsf(k);

        Eigen::Matrix3d JJ = Eigen::Matrix3d::Zero(); // temporary fixed-size matrix to efficiently compute the inverse
        size_t i = 0;
        for (auto it = nodes.begin(); it != nodes.end(); ++it, ++i)
            JJ += dff.col(i) * ((*it)->pos).transpose();
        Js[k] = JJ;
        iJs[k] = JJ.inverse();
        detJs[k] = JJ.determinant();
        this->checkJac(k);
    }
}

/**
 * @brief Compute Jacobian matrix and determinant gradient with respect to each node coordinate
 *
 * dJ[l = i * 3 + j](3,3) = dN_i(3,1) * dxy_i(1,3)/dx_l (for node i and dimension j)
 */
void Tetra4::buildGradientJ()
{
    CacheTetra4 &cache = getCache(order);
    for (size_t k = 0; k < cache.getVGauss().getN(); ++k)
    {
        Eigen::MatrixXd const &dsf = getCache(order).getDsf(k);

        // Jacobian
        gradJs[k].resize(4 * 3); // 4 nodes and 3 dimensions (x, y, z)
        for (size_t i = 0; i < 4; ++i)
        {
            for (size_t j = 0; j < 3; ++j)
            {
                gradJs[k][i * 3 + j] = Eigen::Matrix3d::Zero();
                gradJs[k][i * 3 + j].col(j) = dsf.col(i); // retain sf gradients associated to ith node for dimension j
            }
        }
        // Determinant
        double detJ = this->getDetJ(k);
        Eigen::MatrixXd const &iJ = this->getJinv(k);
        gradDetJs[k].resize(4 * 3);
        for (size_t i = 0; i < 4 * 3; ++i)
            gradDetJs[k][i] = detJ * (iJ * gradJs[k][i]).trace();
    }
}

/**
 * @brief Return the element volume
 */
void Tetra4::computeV()
{
    CacheTetra4 &cache = getCache(order);
    GaussTetra4 &gauss = cache.gauss;

    // Gauss integration
    vol = 0.0;
    for (size_t k = 0; k < gauss.getN(); ++k)
        vol += gauss.getW(k) * this->getDetJ(k);
}
/**
 * @brief Compute volume gradient with respect to each node coordinate
 *
 * dV[i] = w_k * dDetJ[i]_k
 */
void Tetra4::buildGradientV(int dim)
{
    GaussTetra4 &gauss = getCache(order).gauss;

    // Gradient of volume
    gradVol.clear();
    gradVol.resize(4 * dim, 0.); // 4 nodes
    for (size_t k = 0; k < gauss.getN(); ++k)
    {
        double w = gauss.getW(k);                         // GP weight
        std::vector<double> dDetJ = this->getGradDetJ(k); // gradient of Jacobian determinant
        for (size_t i = 0; i < dDetJ.size(); ++i)
            gradVol[i] += w * dDetJ[i];
    }
}

/**
 * @brief Compute gradient of a scalar field at Gauss point k
 */
Eigen::VectorXd Tetra4::computeGradient(std::vector<double> const &u, size_t k) const
{
    // Solution vector
    Eigen::Vector4d ue;
    for (size_t i = 0; i < nodes.size(); ++i)
        ue(i) = u[nodes[i]->row];
    // gradient in global axis: inv(J)_ij * d_jN_i * ue_i
    return this->getJinv(k) * getCache(order).getDsf(k) * ue;
}
